#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
from .libgauss import get_grid,getB,getBm0
from .plotlib import *
from .utils import planetlist, stdDatDir


def getBr(planet, r=1, nphi=256, ntheta=128, info=True):

    p2D,th2D = get_grid(nphi=nphi,ntheta=ntheta)

    if planet.name in ["mercury", "saturn"]:
        Br = getBm0(planet.lmax,
                    planet.glm,
                    r,
                    p2D,
                    th2D) * 1e-3
        dipTheta = 0.
        dipPhi = 0.
    else:
        Br = getB(planet.lmax,
                  planet.glm,
                  planet.hlm,
                  planet.idx,
                  r,
                  p2D,
                  th2D,
                  planet=planet.name) * 1e-3

        dipTheta = np.arctan(np.sqrt(planet.glm[planet.idx[1,1]]**2 + planet.hlm[planet.idx[1,1]]**2)
                                    /planet.glm[planet.idx[1,0]]) * 180./np.pi
        dipPhi = np.arctan(planet.hlm[planet.idx[1,1]]/planet.glm[planet.idx[1,1]]) * 180./np.pi

    if info:
        print(("Planet: %s" %planet.name.capitalize()))
        #print(("Depth (fraction of surface radius) = %.2f" %r))
        print(("l_max = %d" %planet.lmax))
        print(("Dipole tilt (degrees) = %f" %dipTheta))

    return p2D, th2D, Br, dipTheta, dipPhi

def plotAllFields(datDir=stdDatDir,r=1.0,levels=30,cmap='RdBu_r',proj='Mollweide'):

    from .planet import planet as Planet

    print("")
    print('|=========|======|=======|')
    print(('|%-8s | %-2s| %-5s |' %('Planet','Theta','Phi')))
    print('|=========|======|=======|')

    plt.figure(figsize=(12,12))

    for k, name in enumerate(planetlist):
        planet = Planet(name=name,r=r,info=False)

        if name == "ganymede":
            nplot = 8
        else:
            nplot = k+1

        if proj.lower() == 'hammer':
            ax = plt.subplot(3,3,nplot)
        else:
            import cartopy.crs as ccrs
            projection = eval('ccrs.'+proj+'()')
            ax = plt.subplot(3,3,nplot,projection=projection)

        plotB_subplot(planet.p2D,
                      planet.th2D,
                      planet.Br,
                      ax,
                      planet=name,
                      levels=levels,
                      cmap=cmap,
                      proj=proj)

        if name in ["mercury","saturn"]:
            print(('|%-8s | %-4.1f | %-5.1f |' %(name.capitalize(),planet.dipTheta, planet.dipPhi)))
        else:
            print(('|%-8s | %-3.1f | %-5.1f |' %(name.capitalize(),planet.dipTheta, planet.dipPhi)))

    print('|---------|------|-------|')

    if r == 1:
        plt.suptitle(r'Radial magnetic field ($\mu$T) at surface', fontsize=20)
    else:
        plt.suptitle(r'Radial magnetic field ($\mu$T) at $r/r_{\rm surface} = %.2f$' %r, fontsize=20)


def plotMagField(name,r=1,levels=30,proj='moll',cmap='RdBu_r'):
    from .planet import planet as Planet

    planet = Planet(name)
    plt.figure(figsize=(12,6.75))
    plotB(planet.p2D,
          planet.th2D,
          planet.Br,
          planet=planet.name,
          levels=levels,proj=proj,cmap=cmap,r=planet.r)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np

def extrapot(lmax,rcmb,brcmb,rout):

    nphi, ntheta = brcmb.shape
    nrout = len(rout)

    polar_opt = 1e-10

    lmax = int(nphi/3)
    mmax = lmax

    try:
        import shtns
    except ImportError:
        print("Potential extrapolation requires the SHTns library")
        print("It can be obtained here: https://bitbucket.org/nschaeff/shtns")

    norm=shtns.sht_orthonormal | shtns.SHT_NO_CS_PHASE

    sh = shtns.sht(lmax,mmax=mmax,norm=norm)
    ntheta, nphi = sh.set_grid(ntheta, nphi, polar_opt=polar_opt)


    L = sh.l * (sh.l + 1)

    brlm = sh.analys(brcmb.T)
    bpolcmb = np.zeros_like(brlm)
    bpolcmb[1:] = rcmb**2 * brlm[1:]/L[1:]
    btor = np.zeros_like(brlm)

    brout = np.zeros([ntheta,nphi,nrout])
    btout = np.zeros([ntheta,nphi,nrout])
    bpout = np.zeros([ntheta,nphi,nrout])

    for k,radius in enumerate(rout):
        print(("%d/%d" %(k,nrout)))

        radratio = rcmb/radius
        bpol = bpolcmb * radratio**(sh.l)
        brlm = bpol * L/radius**2
        brout[...,k] = sh.synth(brlm)

        dbpoldr = -sh.l/radius * bpol
        slm = dbpoldr

        btout[...,k], bpout[...,k] = sh.synth(slm,btor)

    brout = np.transpose(brout,(1,0,2))
    btout = np.transpose(btout,(1,0,2))
    bpout = np.transpose(bpout,(1,0,2))

    return brout, btout, bpout

def get_grid(r,theta,phi):

    nr,ntheta,nphi = len(r),len(theta),len(phi)

    r3D  = np.zeros([nphi,ntheta,nr])
    th3D = np.zeros([nphi,ntheta,nr])
    p3D  = np.zeros([nphi,ntheta,nr])

    for i in range(nr):
        r3D[...,i] = r[i]
    for j in range(ntheta):
        th3D[:,j,:] = theta[j]
    for k in range(nphi):
        p3D[k,...] = phi[k]

    s = r3D * np.sin(th3D)
    x = s *   np.cos(p3D)
    y = s *   np.sin(p3D)
    z = r3D * np.cos(th3D)

    return r3D,th3D,p3D, x,y,z, s

def get_cart(vr,vt,vp,r3D,th3D,p3D):

    vs = vr * np.sin(th3D) + vt * np.cos(th3D)
    vz = vr * np.cos(th3D) - vt * np.sin(th3D)

    vx = vs * np.cos(p3D) - vp * np.sin(p3D)
    vy = vs * np.sin(p3D) + vp * np.cos(p3D)

    return vx,vy,vz

def writeVts(name,br,btheta,bphi,r,theta,phi):

    r3D,th3D,p3D, x,y,z, s = get_grid(r,theta,phi)

    print("grid shape=",th3D.shape)

    bx,by,bz = get_cart(br, btheta, bphi,r3D,th3D,p3D)

    try:
        try: # Version 2 changed naming convention of functions
            from evtk.hl import structuredToVTK
            gridToVTK = structuredToVTK
        except:
            import evtk
            gridToVTK = evtk.hl.gridToVTK
    except:
        try:
            from pyevtk.hl import gridToVTK
        except:
            print("This requires the use of evtk library!")
            print("You can get it from https://github.com/paulo-herrera/PyEVTK")


    br = np.asfortranarray(br)
    bx = np.asfortranarray(bx)
    by = np.asfortranarray(by)
    bz = np.asfortranarray(bz)

    gridToVTK("%s"%name,x,y,z,pointData= {"radius":r3D,
                                          "Radial mag field":br,
                                          "Mag Field":(bx, by,bz)})


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import numpy as np
import matplotlib.pyplot as plt
from .libgauss import get_data, filt_Gauss, filt_Gaussm0,getB, getBm0, get_spec
from .libbfield import getBr
from .plotlib import plotSurf, plot_spec
from .utils import stdDatDir, planetlist


class planet:

    def __init__(self,name='earth',r=1,nphi=256,datDir=stdDatDir,info=True):

        self.name   = name.lower()
        self.nphi   = nphi
        self.ntheta = nphi//2

        if self.name not in planetlist:
            print("Planet must be one of the following!")
            print(planetlist)

        self.datDir = datDir
        self.glm, self.hlm, self.lmax, self.idx = \
                get_data(self.datDir,planet=self.name)

        self.p2D, self.th2D, self.Br, self.dipTheta, self.dipPhi = \
                getBr(self,r=r,
                    nphi=self.nphi,
                    ntheta=self.ntheta,
                    info=info)

        self.phi = self.p2D[:,0]
        self.theta = self.th2D[0,:]
        self.r = r

    def plot(self,r=1,levels=30,cmap='RdBu_r',proj='Mollweide'):
        plt.figure(figsize=(12,6.75))

        if r == 1:
            ax,cbar = plotSurf(self.p2D,self.th2D,self.Br,levels=levels,cmap=cmap,proj=proj)
        else:
            self.p2D, self.th2D, self.Br, self.dipTheta, self.dipPhi = \
                    getBr(planet=self,r=r,info=False)
            self.r = r
            ax,cbar = plotSurf(self.p2D,self.th2D,self.Br,levels=levels,cmap=cmap,proj=proj)

        cbar.ax.set_xlabel(r'Radial magnetic field ($\mu$T)',fontsize=25)
        cbar.ax.tick_params(labelsize=20)

        if r==1:
            radLabel = '  Surface'
        else:
            radLabel = r'  $r/r_{\rm surface}=%.2f$' %r

        if proj.lower() != 'hammer' and self.name == 'earth':
            ax.coastlines()

        if r==1:
            radLabel = '  Surface'
        else:
            radLabel = r'  $r/r_{\rm surface}=%.2f$' %r

        ax.set_title(self.name.capitalize() + radLabel,fontsize=25,pad=20)
        plt.tight_layout()

    def writeVtsFile(self,potExtra=False,ratio_out=2,nrout=32):
            from .potextra import extrapot, writeVts

            rout = np.linspace(1,ratio_out,nrout)
            if potExtra:
                brout, btout, bpout = extrapot(self.lmax,1.,self.Br,rout)
            else:
                brout = self.Br
                btout = np.zeros_like(self.Br)
                bpout = np.zeros_like(self.Br)

            writeVts(self.name,brout,btout,bpout,rout,self.theta,self.phi)

    ## Filtered plots

    def plot_filt(self,r=1,larr=None,marr=None,lCutMin=0,lCutMax=None,mmin=0,mmax=None,levels=30,cmap='RdBu_r',proj='Mollweide'):

        self.larr_filt = larr
        self.marr_filt = marr
        self.lCutMin = lCutMin
        self.lCutMax = lCutMax
        self.mmin_filt = mmin
        self.mmax_filt = mmax

        if self.lCutMax is None:
            self.lCutMax = self.lmax
        if self.mmax_filt is None:
            self.mmax_filt = self.lmax

        self.r_filt = r

        if self.name in ['mercury','saturn']:
            self.glm_filt,self.hlm_filt =\
                    filt_Gaussm0(self.glm,self.hlm,self.lmax,larr=self.larr_filt,\
                        lCutMin=self.lCutMin,lCutMax=self.lCutMax)
            self.Br_filt = 1e-3*getBm0(self.lmax,self.glm_filt,self.r_filt,self.p2D,self.th2D)
        else:
            self.glm_filt,self.hlm_filt =\
                filt_Gauss(self.glm,self.hlm,self.lmax,self.idx,larr=self.larr_filt,\
                    marr=self.marr_filt,lCutMin=self.lCutMin,lCutMax=self.lCutMax,mmin=self.mmin_filt,mmax=self.mmax_filt)

            self.Br_filt = 1e-3*getB(self.lmax,self.glm_filt,self.hlm_filt,self.idx,self.r_filt,self.p2D,self.th2D,planet=self.name)

        plt.figure(figsize=(12,6.75))

        ax,cbar = plotSurf(self.p2D,self.th2D,self.Br_filt,levels=levels,cmap=cmap,proj=proj)

        if r==1:
            radLabel = '  Surface'
        else:
            radLabel = r'  $r/r_{\rm surface}=%.2f$' %r

        if self.larr_filt is not None:
            elllabel = r', $l = %s$' %np.str(self.larr_filt)
        else:
            if self.lCutMin > 0:
                if self.lCutMax < self.lmax:
                    elllabel = r', $ %d \leq l \leq %d$' %(self.lCutMin,self.lCutMax)
                else:
                    elllabel = r', $l \geq %d$' %self.lCutMin

            elif self.lCutMax < self.lmax:
                elllabel = r', $l \leq %d$' %self.lCutMax

        if self.marr_filt is not None:
            elllabel += r', $m = %s$' %np.str(self.marr_filt)
        else:
            if self.mmin_filt > 0:
                if self.mmax_filt < self.lmax:
                    elllabel += r', $ %d \leq m \leq %d$' %(self.mmin_filt,self.mmax_filt)
                else:
                    elllabel += r', $m \geq %d$' %self.mmin_filt
            elif self.mmax_filt < self.lmax:
                elllabel += r', $m \leq %d$' %self.mmax_filt

        cbar.ax.set_xlabel(r'Radial magnetic field ($\mu$T)',fontsize=25)
        cbar.ax.tick_params(labelsize=20)

        if proj.lower() != 'hammer' and self.name == 'earth':
            ax.coastlines()
        ax.set_title(self.name.capitalize() + radLabel + elllabel,fontsize=25,pad=20)
        plt.tight_layout()


    def spec(self,r=1,iplot=True):
        self.emag_spec, emag_10 = get_spec(self.glm,self.hlm,self.idx,self.lmax,planet=self.name,r=r)
        l = np.arange(self.lmax+1)

        self.dip_tot = self.emag_spec[1]/sum(self.emag_spec)
        self.dipolarity = emag_10/sum(self.emag_spec)
        if iplot:
            plt.figure(figsize=(7,7))
            plot_spec(l,self.emag_spec,r,self.name)
            plt.tight_layout()
            plt.show()


import numpy as np
from astroedu.constants import G, au

class Body2D:
    """Body object"""

    def __init__(self, name, x, y, r, m):
        """Initialises the body

        Args:
            name -- string, name of the body
            x    -- float, x position of the object (AU) - is converted to m
            y    -- float, y position of the object (AU) - is converted to m
            r    -- float, radius of the object (km) - is converted to m
            m    -- float, mass of the object (kg)

        Example:
            >>> moon = Body2D('Moon', 0, 0, 1737.4, 0.07346*10**24)
        """
        self.name = name
        self.x = x*au
        self.y = y*au
        self.r = r*1000
        self.m = m

    def __str__(self):
        return f'{self.name} at ({self.x/au:.2f}, {self.y/au:.2f}) AU with r = {self.r/1000:.2E} km and m = {self.m:.2E} kg'

    def __repr__(self):
        return f'Body({self.name}, {self.x/au:f}, {self.y/au:f}, {self.r/1000:f}, {self.m:f})'

    @classmethod
    def Sun(cls, x, y):
        return cls('Sun', x, y, 695700, 1988500*10**24)

    @classmethod
    def Earth(cls, x, y):
        return cls('Earth', x, y, 6371, 5.9724*10**24)

    @classmethod
    def Moon(cls, x, y):
        return cls('Moon', x, y, 1737.4, 0.07346*10**24)

    def distance_to(self, body):
        ''' Returns distance between two Body objects in AU
        '''
        x_diff = body.x - self.x
        y_diff = body.y - self.y
        return np.hypot(x_diff, y_diff)

    def angle_to(self, body):
        ''' Returns the angle offset from +x axis counterclockwise in radians 
        '''
        x_diff = body.x - self.x
        y_diff = body.y - self.y
        ang = np.arctan2(y_diff, x_diff)
        if ang >= 0:
            return ang
        else:
            return 2*np.pi + ang

    def force_centre(self, body):
        """ Calculates the gravitational attraction on the main object
        due to the other body. Unit: N

        Args:
            body -- Body class object

        Returns:
            force_c -- float, the gravitational force at the centre of the Main body due to body

        Example:
            >>>  force_c = earth.force_centre(moon)
            >>>  print(force_c)
            3.31874952061913e-05
        """
        force_c = G*self.m*body.m/self.distance_to(body)**2 / self.scale
        return force_c

    def force_points(self, body):
        """ Calculates the gravitational attraction at points on the surface
        of the Main body due to the other body. Unit: N

        Args:
            body -- Body class object

        Returns:
            force_p_h -- NumPy array of floats, hor. component of grav force at each surface point
            force_p_v -- NumPy array of floats, vert. component of grav force at each surface point

        Example:
            >>>  force_p_h, force_p_v = earth.force_points(moon)
            >>>  print(force_p_h)
            array([3.43155527e-05, 3.39721024e-05, 3.31738253e-05, 3.24165524e-05,
                   3.21141611e-05, 3.24165524e-05, 3.31738253e-05, 3.39721024e-05])
        """
        ang = self.angle_to(body)
        dist = self.distance_to(body)
        c = np.sqrt(self.r**2 + dist**2 - 2*self.r*dist*np.cos(self.thetas-ang))
        force_p = G*self.m*body.m/c**2 / self.scale
        sin_b = (dist - self.r*np.cos(self.thetas-ang))/c
        b = np.arcsin(sin_b)
        force_p_h = force_p*np.cos(b+ang-np.pi/2) 
        force_p_v = force_p*np.sin(b+ang-np.pi/2)
        # fix the signs of our vectors
        if abs(dist*np.cos(ang)) < self.r:
            signs_h = np.sign(-np.cos(self.thetas))
            force_p_h = np.absolute(force_p_h)*signs_h
            sign_v = np.sign(np.sin(ang))
            force_p_v = np.absolute(force_p_v)*sign_v
        elif abs(dist*np.sin(ang)) < self.r:
            signs_v = np.sign(-np.sin(self.thetas))
            force_p_v = np.absolute(force_p_v)*signs_v
            sign_h = np.sign(np.cos(ang))
            force_p_h = np.absolute(force_p_h)*sign_h
        else:
            sign_h = np.sign(np.cos(ang))
            force_p_h = np.absolute(force_p_h)*sign_h
            sign_v = np.sign(np.sin(ang))
            force_p_v = np.absolute(force_p_v)*sign_v
        return force_p_h, force_p_v

    def forces(self, body):
        """ Calculates the tidal force at each surface point.

        Args:
            body -- Body class object

        Returns:
            A list containing:
            force_c -- float, the gravitational force at the centre of the Main body due to body
            force_p_h -- NumPy array of floats, hor. component of grav force at each surface point
            force_p_h_diff -- NumPy array of floats, hor. comp. of tidal force at each surface point
            force_p_v -- NumPy array of floats, vert. component of grav force at each surface point
            force_p_v_diff -- NumPy array of floats, vert. comp. of tidal force at each surface point

        Example:
            >>>  t_forces = earth.forces(moon)
            >>>  print(t_forces)
            [3.31874952061913e-05, array([3.43155527e-05, 3.39721024e-05, 3.31738253e-05, 3.24165524e-05,
             3.21141611e-05, 3.24165524e-05, 3.31738253e-05, 3.39721024e-05]), array([ 1.12805754e-06,  7.84607232e-07,  0.00000000e+00, -7.70942821e-07,
             -1.07333414e-06, -7.70942821e-07,  0.00000000e+00,  7.84607232e-07]), array([ 0.00000000e+00, -4.02857477e-07, -5.49819045e-07, -3.75505178e-07,
             -0.00000000e+00,  3.75505178e-07,  5.49819045e-07,  4.02857477e-07]), array([ 0.00000000e+00, -4.02857477e-07, -5.49819045e-07, -3.75505178e-07,
             0.00000000e+00,  3.75505178e-07,  5.49819045e-07,  4.02857477e-07])]
        """
        force_c = self.force_centre(body)
        force_p_h, force_p_v = self.force_points(body)
        ang = self.angle_to(body)
        force_p_h_diff = force_p_h - force_c*np.cos(ang)
        force_p_v_diff = force_p_v - force_c*np.sin(ang)
        zero_cut = 0
        force_p_h_diff[np.abs(force_p_h_diff) < zero_cut] = 0
        force_p_v_diff[np.abs(force_p_v_diff) < zero_cut] = 0
        return [force_c, force_p_h, force_p_h_diff, force_p_v, force_p_v_diff]

    def update_tides(self, key1, key2, off, offsets,  tide_vals=None, use_prev=False):
        """ Updates the tidal forces contained in the tide_f dictionary.

        Called by self.tides() - should not be called independently
        If use_prev is passed as True then tide_vals is not used

        Use previously calculated values for a body if they are already in the dictionary tide_f
        Otherwise update the tide values for the combined effect of all the bodies
        using the passed tide_vals list which is from self.forces()

        Args:
            key1 -- str, the final key of all tidal effects combined, eg 'Sun+Moon'
            key2 -- str, name of the body object, eg 'Sun'
            off -- str, offset_angle of the Body object as a string
            offsets -- str, offset angles of all Body object combined
                    -- eg, '0+0' for two Bodys with offset angles of 0
            tide_vals -- NumPy array of floats from self.forces(), default None
                      -- Is not neccessary when using use_prev=True
            use_prev -- boolean, whether to use values already calculated in tide_f, default False

        Example:
            self.update_tides(tide_f_key, body.name, this_off, tide_f_offs, use_prev=True)
            self.update_tides(tide_f_key, body.name, this_off, tide_f_offs, tide_vals=self.forces(body))
        """
        # check if the final key exists
        if key1 not in self.tide_f:
            self.tide_f[key1] = {}
            self.tide_f[key1][offsets] = []
        # if offsets key not in nested dictionary add it 
        elif offsets not in self.tide_f[key1]:
            self.tide_f[key1][offsets] = []
        # use prev calculated values
        if use_prev:
            # if the final key is empty zip won't work!
            if self.tide_f[key1][offsets] != []:                         
                update = [new_tide+old_tide for new_tide, old_tide in zip(self.tide_f[key2][off], self.tide_f[key1][offsets])]
                self.tide_f[key1][offsets] = update
            else:
                self.tide_f[key1][offsets] = self.tide_f[key2][off]
        # use newly calculated values
        else:
            # add key for single body and its offset
            self.tide_f[key2] = {}
            self.tide_f[key2][off] = tide_vals
            # If there is only 1 body the keys are the same so don't double numbers  
            if key1!=key2: 
                # if the final key is empty zip won't work!
                if self.tide_f[key1][offsets] != []:                         
                    update = [new_tide+old_tide for new_tide, old_tide in zip(tide_vals, self.tide_f[key1][offsets])]
                    self.tide_f[key1][offsets] = update
                else:
                    self.tide_f[key1][offsets] = tide_vals

    def tides(self, *args, step=0.25, scale=1):
        """Calculates the tides on Main due to the bodies passed as arguments.

        Checks to see whether the tidal force from each individual body has been
        calculated previously so it doesn't repeat calculations

        Args:
            step -- float, angular step to make when calculating surface points - in terms of pi
                 -- used to calculate the angles of each point, each point is +step*np.pi
            scale -- float, scale to divide forces by in kg
                  -- set to mass of Main object for acceleration in ms-2
            Any number of Body class objects can then be passed.

        Returns:
            A list containing:
            force_c -- float, the gravitational force at the centre of the Main body due to bodies
            force_p_h -- NumPy array of floats, hor. component of grav force at each surface point
            force_p_h_diff -- NumPy array of floats, hor. comp. of tidal force at each surface point
            force_p_v -- NumPy array of floats, vert. component of grav force at each surface point
            force_p_v_diff -- NumPy array of floats, vert. comp. of tidal force at each surface point

        Example:
            >>> forces = earth.tides(moon, step=0.25, scale=5.972*10**24)
            >>> print(forces)
            [3.31874952061913e-05, array([3.43155527e-05, 3.39721024e-05, 3.31738253e-05, 3.24165524e-05,
             3.21141611e-05, 3.24165524e-05, 3.31738253e-05, 3.39721024e-05]), array([ 1.12805754e-06,  7.84607232e-07,  0.00000000e+00, -7.70942821e-07,
             -1.07333414e-06, -7.70942821e-07,  0.00000000e+00,  7.84607232e-07]), array([ 0.00000000e+00, -4.02857477e-07, -5.49819045e-07, -3.75505178e-07,
             -0.00000000e+00,  3.75505178e-07,  5.49819045e-07,  4.02857477e-07]), array([ 0.00000000e+00, -4.02857477e-07, -5.49819045e-07, -3.75505178e-07,
             0.00000000e+00,  3.75505178e-07,  5.49819045e-07,  4.02857477e-07])]
        """
        self.step = step              # in terms of pi
        self.thetas = np.arange(0, 2*np.pi, step*np.pi)
        self.scale = scale
        self.tide_f = {

        }
        # get all body names then join them to make the dictionary key
        body_names = [body.name for body in args]
        tide_f_key = "+".join(body_names)
        # get all offset angles then join them to make the nested dictionary key
        list_offsets = [str(self.angle_to(body)) for body in args]
        tide_f_offs = "+".join(list_offsets)
        for body in args:
            this_off = str(self.angle_to(body))
            # if the tidal force for this body has been calculated before use stored values
            if body.name in self.tide_f:
                # check if the calculated values were for the same offset_angle
                if this_off in self.tide_f[body.name]:
                    self.update_tides(tide_f_key, body.name, this_off, tide_f_offs, use_prev=True)
            # else calculate the tidal force from the new body and pass it to self.update_tides()
            else:
                this_body_tide = self.forces(body)
                self.update_tides(tide_f_key, body.name, this_off, tide_f_offs, tide_vals=this_body_tide)
        return self.tide_f[tide_f_key][tide_f_offs]


import numpy as np

class Constant:
    ''' Class for defining all constants

        For more functionality use astropy's constant instead
        https://docs.astropy.org/en/stable/constants/
        Most of the constants were taken from the page above

    Example:
        >>> c = Constant('c', 299792458, 'm/s', 'Speed of light')
        >>> c
        299792458
    '''

    def __init__(self, symbol, value, unit, name):
        self.symbol = symbol
        self.value = value
        self.unit = unit 
        self.name = name

    def __repr__(self):
        return f'Constant({self.symbol}, {self.value}, {self.unit}, {self.name})'

    def __str__(self):
        return f'Name = {self.name}\nValue = {self.value}\nUnit = {self.unit}'

    def __add__(self, other):
        if isinstance(other, Constant):
            return self.value + other.value
        else:
            return self.value + other

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, Constant):
            return self.value - other.value
        else:
            return self.value - other

    def __rsub__(self, other):
        return other - self.value

    def __mul__(self, other):
        if isinstance(other, Constant):
            return self.value * other.value
        else:
            return self.value * other

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, Constant):
            return self.value / other.value
        else:
            return self.value / other

    def __rtruediv__(self, other):
        return other / self.value

    def __pow__(self, other):
        if isinstance(other, Constant):
            return self.value ** other.value
        else:
            return self.value ** other

    def __rpow__(self, other):
        return other ** self.value

atm = Constant('atm', 101325, 'Pa', 'Standard Atmosphere')

au = Constant('AU', 1.49597871e+11, 'm', 'Astronomical Unit')

b_wien = Constant('b', 2.897771955e-3, 'm K', 'Wiens Displacement constant')

c = Constant('c', 299792458, 'm/s', 'Speed of light')

e = Constant('e', 1.60217663e-19, 'C', 'Electron Charge')

g0 = Constant('g', 9.80665, 'm/s2', 'Standard acceleration of gravity')

G = Constant('G', 6.6743e-11, 'm3 / (kg s2)', 'Gravitational constant')

h = Constant('h', 6.62607015e-34, 'Js', 'Planck constant')

hbar = Constant('hbar', 1.05457182e-34, 'Js', 'Reduced Planck constant')

k_B = Constant('kB', 1.380649e-23, 'J/K', 'Boltzman constant')

L_bol0 = Constant('L_bol0', 3.0128e+28, 'W', 'Luminosity for absolute bolometric magnitude 0')

L_sun = Constant('L_sun', 3.828e+26, 'W', 'Nominal Solar Luminosity')

m_e = Constant('m_e', 9.1093837e-31, 'kg', 'Electron mass')

m_n = Constant('m_n', 1.6749275e-27, 'kg', 'Neutron mass')

m_p = Constant('m_p', 1.67262192e-27, 'kg', 'Proton mass')

M_earth = Constant('M_earth', 5.97216787e+24, 'kg', 'Earth mass')

M_jup = Constant('M_jup', 1.8981246e+27, 'kg', 'Jupiter mass')

M_sun = Constant('M_sun', 1.98840987e+30, 'kg', 'Solar mass')

pc = Constant('pc', 3.08567758e+16, 'm', 'Parsec')

R_earth = Constant('R_earth', 5.97216787e+24, 'm', 'Nominal Earth equatorial radius')

R_jup = Constant('R_jup', 1.8981246e+27, 'm', 'Nominal Jupiter equatorial radius')

R_sun = Constant('R_sun', 1.98840987e+30, 'm', 'Nominal solar radius')

sigma_sb = Constant('sigma_sb', 5.67037442e-08, 'W / (K4 m2)', 'Stefan-Boltzmann constant')

u = Constant('u', 1.66053907e-27, 'kg', 'Atomic mass unit')


import subprocess
from importlib_resources import files, is_resource
from importlib.util import find_spec
from configparser import ConfigParser
import fire

from astroedu.__build__ import get_astroedu_path
from astroedu.__build__ import build_path_config

commands = ['build', 'interactive', 'magfield']
astroedu_path = get_astroedu_path()
config_file_name = '/config.ini'
config_file_path = astroedu_path + config_file_name

def main():
    ''' Entry point for package
    '''
    fire.Fire(startup)

def load_astroedu_config():
    ''' Loads the config.ini file

    Returns:
        The config dictionary

    Example:
        >>> astroedu_config = load_astroedu_config()
    '''
    if not is_resource('astroedu', 'config.ini'):
        raise FileExistsError('No config.ini file found. Did you forget to run: \n astroedu build')
    config = ConfigParser()
    config.read(config_file_path)
    return config['astroedu-config']

def _load_interactive(*args):
    ''' Loads an interactive notebook
        Called from startup() only
    '''
    interactive_name = args[1]
    if not is_resource('astroedu.interactives', interactive_name+'.ipynb'):
        raise FileExistsError(f'No interactive named {interactive_name} found.')
    with files('astroedu').joinpath('interactives', interactive_name+'.ipynb') as p:
        call = ['jupyter', 'lab']
        call_args = list(args[2:]) if len(args) > 2 else []
        run_interactive = call + [str(p)] + call_args
        print(f'Loading interactive {interactive_name}!')
        subprocess.run(run_interactive)

def _planetmagfield_quick(*args):
    ''' Utility function to make wuick plots using planetMagFields
    See https://zenodo.org/record/5140421#.YSdp7t_TUuU
    '''
    levels=20
    cmap='RdBu_r'
    proj = 'Mollweide'
    r = 1

    import numpy as np
    
    if len(args) > 4:
        raise ValueError("Too many arguments, exiting ...\n")
    elif len(args) == 4:
        planet = str(args[1]).lower()
        r      = np.float32(args[2])
        proj   = str(args[3])
    elif len(args) == 3:
        planet = str(args[1]).lower()
        try:
            r      = np.float32(args[2])
        except:
            proj   = str(args[2])
    elif len(args) == 2:
        if args[2] == '--help':
            print('planetMagFields\nSubmodule by Ankit Barik - 10.5281/zenodo.5140421\nUsage see - github.com/AnkitBarik/planetMagFields')
        else:
            print("Radius not specified, using surface\n")
            planet = str(args[1]).lower()
    else:
        print("Planet or radius not specified, plotting for Earth's surface\n")
        planet="earth"
        r=1.

    import matplotlib.pyplot as plt
    # from astroedu.planetmagfields.libbfield import getBr, plotAllFields, plotMagField

    if planet == 'all':
        from astroedu.planetmagfields.libbfield import plotAllFields

        with files('astroedu.planetmagfields').joinpath('data/') as p:
            plotAllFields(datDir=p,r=r,levels=levels,cmap=cmap,proj=proj)
            plt.tight_layout()
            plt.subplots_adjust(top=0.895,
                                bottom=0.035,
                                left=0.023,
                                right=0.976,
                                hspace=0.38,
                                wspace=0.109)
    else:
        from astroedu.planetmagfields.libbfield import plotMagField

        with files('astroedu.planetmagfields').joinpath('data/') as p:
            # plotMagField(planet=planet,r=r,datDir=p,levels=levels,cmap=cmap,proj=proj)
            plotMagField(name=planet,r=r,levels=levels,cmap=cmap,proj=proj)
            plt.tight_layout()

    plt.show()

def startup(*args):
    ''' Startup function called by Fire cli function main()

    Based on arguments passed to astroedu call
    - builds config.ini file
    - launches interactive notebook

    Args:
        args[0] -- string, name of astroedu command to call

    Example:
        Loads the wiens_law interactive and launches jupyterlab
        >>> astroedu interactive wiens_law
        Arguments passed after the interactive name go to jupyter
        >>> astroedu interactive wiens_law --port 9999

        To build the config.ini file
        >>> astroedu build
    '''
    if not args:
        raise ValueError('No arguments passed to astroedu call.')
    command = args[0].lower()
    if command not in commands:
        raise ValueError('Command not recognised.')
    elif command == 'build': 
        build_path_config()
    elif len(args) == 1: # build doesn't need extras
        raise ValueError(f'No arguments passed to {command}')
    elif command == 'interactive':
        _load_interactive(*args)
    elif command == 'magfield':
        _planetmagfield_quick(*args)

if __name__ == '__main__':
    fire.Fire(startup)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors

def hammer2cart(ttheta, pphi, colat=False):
    """
    This function is used to define the Hammer projection for
    default plotting. Copied from MagIC python plotting script:

    https://github.com/magic-sph/magic/blob/master/python/magic/plotlib.py
    """

    if not colat: # for lat and phi \in [-pi, pi]
        xx = 2.*np.sqrt(2.) * np.cos(ttheta)*np.sin(pphi/2.)\
             /np.sqrt(1.+np.cos(ttheta)*np.cos(pphi/2.))
        yy = np.sqrt(2.) * np.sin(ttheta)\
             /np.sqrt(1.+np.cos(ttheta)*np.cos(pphi/2.))
    else:  # for colat and phi \in [0, 2pi]
        xx = -2.*np.sqrt(2.) * np.sin(ttheta)*np.cos(pphi/2.)\
             /np.sqrt(1.+np.sin(ttheta)*np.sin(pphi/2.))
        yy = np.sqrt(2.) * np.cos(ttheta)\
             /np.sqrt(1.+np.sin(ttheta)*np.sin(pphi/2.))
    return xx, yy

def plotSurf(p2D,th2D,B,levels=60,cmap='RdBu_r',proj='Mollweide'):

    bmax = np.abs(B).max()
    digits = int(np.log10(bmax)) + 1

    if digits > 1:
        bmax = np.round(bmax)
    else:
        bmax = np.round(bmax,decimals=1)

    divnorm = colors.TwoSlopeNorm(vmin=-bmax, vcenter=0, vmax=bmax)
    cs = np.linspace(-bmax,bmax,levels)


    lon2D = p2D - np.pi
    lat2D = np.pi/2 - th2D

    try:
        import cartopy.crs as ccrs
    except:
        print("cartopy library not available, using Hammer projection")
        proj = 'hammer'

    if proj.lower() == 'hammer':
        ax = plt.axes()
        xx,yy = hammer2cart(lat2D,lon2D)
        cont = ax.contourf(xx,yy,B,cs,cmap=cmap,norm=divnorm,extend='both')
    else:
        projection = eval('ccrs.'+proj+'()')

        ax = plt.axes(projection=projection)

        cont = ax.contourf(lon2D*180/np.pi,lat2D*180/np.pi,B,cs,  \
            transform=ccrs.PlateCarree(),cmap=cmap,norm=divnorm,extend='both')

    cbar = plt.colorbar(cont,orientation='horizontal',fraction=0.06, pad=0.04,ticks=[-bmax,0,bmax])

    ax.axis('equal')
    ax.axis('off')

    return ax, cbar

def plotB(p2D,th2D,B,r=1,planet="earth",levels=60,cmap='RdBu_r',proj='Mollweide'):

    planet = planet.lower()

    ax,cbar = plotSurf(p2D,th2D,B,levels=levels,cmap=cmap,proj=proj)
    cbar.ax.set_xlabel(r'Radial magnetic field ($\mu$T)',fontsize=25)
    cbar.ax.tick_params(labelsize=20)

    if proj.lower() != 'hammer' and planet == 'earth':
        ax.coastlines()

    if r==1:
        radLabel = '  Surface'
    else:
        radLabel = r'  $r/r_{\rm surface}=%.2f$' %r

    ax.set_title(planet.capitalize() + radLabel,fontsize=25,pad=20)

def plotB_subplot(p2D,th2D,B,ax,planet="earth",levels=60,cmap='RdBu_r',proj='Mollweide'):
    planet = planet.lower()

    bmax = np.abs(B).max()
    digits = int(np.log10(bmax)) + 1

    if digits > 1:
        bmax = np.round(bmax)
    else:
        bmax = np.round(bmax,decimals=1)

    p2D -= np.pi
    th2D -= np.pi/2
    th2D = -th2D

    cs = np.linspace(-bmax,bmax,levels)
    divnorm = colors.TwoSlopeNorm(vmin=-bmax, vcenter=0, vmax=bmax)

    try:
        import cartopy.crs as ccrs
    except:
        print("cartopy library not available, using Hammer projection")
        proj = 'hammer'

    if proj.lower() == 'hammer':
        xx,yy = hammer2cart(th2D,p2D)
        cont = ax.contourf(xx,yy,B,cs,cmap=cmap,norm=divnorm,extend='both')
    else:
        if planet == "earth":
            ax.coastlines()

        cont = ax.contourf(p2D*180/np.pi,th2D*180/np.pi,B,cs,  \
            transform=ccrs.PlateCarree(),cmap=cmap,norm=divnorm,extend='both')

    cbar = plt.colorbar(cont,orientation='horizontal',fraction=0.06, pad=0.04,ticks=[-bmax,0,bmax])
    cbar.ax.tick_params(labelsize=15)

    ax.set_title(planet.capitalize(),fontsize=20)
    ax.axis('equal')
    ax.axis('off')

def plot_spec(l,E,r,planet):

    plt.semilogy(l[1:],E[1:],'-o',lw=1.2,color='#449c99',mfc='#347b79',mec='#347b79',ms=8)
    plt.xlabel(r'$l$',fontsize=30)
    plt.ylabel(r'$R_l$ (nT$^2$)',fontsize=30)
    plt.xticks(l)

    plt.tick_params(which='major',labelsize=15,length=10)
    plt.tick_params(which='minor',length=5)

    plt.grid(True,alpha=0.5)

    if r==1:
        radLabel = '  Surface'
    else:
        radLabel = r'  $r/r_{\rm surface}=%.2f$' %r

    plt.title('Lowes spectrum, '+ planet.capitalize() + radLabel,fontsize=20,pad=10)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# import os
from importlib_resources import files

# stdDatDir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data/')
stdDatDir = files('astroedu.planetmagfields').joinpath('data/')

planetlist = ["mercury", "earth", "jupiter", "saturn", "uranus", "neptune",
              "ganymede"]


import numpy as np
from astroedu import constants as const

def wiens_law(a):
    ''' Computes the peak wavelengths if temperatures are passed
        Computes the temperatures if peak wavelengths are passed

    0 must not be passed as an argument since 1/0 will fail

    Args:
        a -- float or 1D np.ndarray, peak wavelength or temperature

    Returns:
        float or 1D np.ndarray, peak wavelength or temperature

    Example:
        To get peak wavelengths pass the temperatures to arg a
        >>> T = np.arange(0.1, 3000, 10) # K
        >>> wiens_law(T)
        To get a temperature pass a wavelength
        >>> l = 500*10**(-9) # m
        >>> wiens_law(l)
    '''
    return const.b_wien/a

def plancks_law(T, l):
    ''' Computes the blackbody curve over wavelength range l for temp T
        using Planck's law

    Args:
        T -- float, temperature of blackbody
        l -- float or 1D np.ndarray, wavelengths

    Returns:
        float or 1D np.ndarray, intensity in units: 10^4 W sr^-1 m^-2 nm^-1

    Example:
        >>> plancks_law(300, 560*10**-9)
    '''
    return 2*const.h*const.c**2/l**5 * 1/(np.exp(const.h*const.c/(l*const.k_B*T))-1)


'''
astroedu.datasets - utilities for loading datasets
'''

from importlib_resources import files, is_resource
import pandas as pd

help_text = '''Pass the name of a dataset to load_data as a string.\n
For instance to load the planets dataset run load_data('planets').\n
The returned value is a Pandas dataframe.\n
To see all available datasets go to https://github.com/astroDimitrios/astroedu/tree/main/src/astroedu/datasets'''

def load_data(data, info=False):
    ''' Loads a dataset as a Pandas dataframe

    Special case data == 'help' the help_text is printed to console
    All datasets visible at:
    https://github.com/astroDimitrios/astroedu/tree/main/src/astroedu/datasets

    Args:
        data -- string, name of the csv file to load (without extension)
        info -- boolean, default False - if True prints a short description
                of the data before loading the dataframe

    Returns:
        A pandas dataframe

    Example:
        >>> planets = load_data('planets')
        >>> load_data('help')
    '''
    if data == 'help':
        print(help_text)
        return
    data_csv = data+'.csv'
    if not is_resource('astroedu.datasets', data_csv):
            raise FileExistsError(f'No data file named {data_csv} found.')
    with files('astroedu.datasets').joinpath(data_csv) as p:
        if info:
            with files('astroedu.datasets').joinpath(data+'.txt') as i:
                InfoHandler = open(str(i), 'r')
                print(InfoHandler.read())
        return pd.read_csv(str(p))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
from .libgauss import get_data,gen_idx,get_grid,getB,getBm0
from .plotlib import *
from importlib_resources import files
import sys

planetlist = ["mercury","earth","jupiter","saturn","uranus","neptune","ganymede"]

datDir = files('astroedu.planetmagfields').joinpath('data/')

def getBr(datDir=datDir,planet="earth",r=1,info=True):

    if planet not in planetlist:
        print("Planet must be one of the following!")
        print(planetlist)
        sys.exit()

    if planet == '--help':
        print("Usage: ./magField.py <planet name>")
        print("Example: ./magField.py earth")
        sys.exit()

    g,h,lmax,idx = get_data(datDir,planet)

    p2D,th2D = get_grid()

    if planet in ["mercury","saturn"]:
        Br = getBm0(lmax,g,r,p2D,th2D) * 1e-3
        dipTheta = 0.
        dipPhi = 0.
    else:
        Br = getB(lmax,g,h,idx,r,p2D,th2D,planet=planet) * 1e-3

        dipTheta = np.arctan(np.sqrt(g[idx[1,1]]**2 + h[idx[1,1]]**2)/g[idx[1,0]]) * 180./np.pi
        dipPhi = np.arctan(h[idx[1,1]]/g[idx[1,1]]) * 180./np.pi

    if info:
        print(("Planet: %s" %planet.capitalize()))
        #print(("Depth (fraction of surface radius) = %.2f" %r))
        print(("l_max = %d" %lmax))
        print(("Dipole tilt (degrees) = %f" %dipTheta))

    return p2D, th2D, Br, dipTheta, dipPhi

def plotAllFields(datDir=datDir,r=1.0,levels=30,cmap='RdBu_r',proj='Mollweide'):

    print("")
    print('|=========|======|=======|')
    print(('|%-8s | %-2s| %-5s |' %('Planet','Theta','Phi')))
    print('|=========|======|=======|')

    plt.figure(figsize=(12,12))
    for k, planet in enumerate(planetlist):
        p2D,th2D,Br,dipTheta,dipPhi = getBr(datDir=datDir,planet=planet,r=r,info=False)

        if planet == "ganymede":
            nplot = 8
        else:
            nplot = k+1

        if proj.lower() == 'hammer':
            ax = plt.subplot(3,3,nplot)
        else:
            import cartopy.crs as ccrs
            projection = eval('ccrs.'+proj+'()')
            ax = plt.subplot(3,3,nplot,projection=projection)

        plotB_subplot(p2D,th2D,Br,ax,planet=planet,levels=levels,cmap=cmap,proj=proj)

        if planet in ["mercury","saturn"]:
            print(('|%-8s | %-4.1f | %-5.1f |' %(planet.capitalize(),dipTheta, dipPhi)))
        else:
            print(('|%-8s | %-3.1f | %-5.1f |' %(planet.capitalize(),dipTheta, dipPhi)))

    print('|---------|------|-------|')

    if r == 1:
        plt.suptitle(r'Radial magnetic field ($\mu$T) at surface', fontsize=20)
    else:
        plt.suptitle(r'Radial magnetic field ($\mu$T) at $r/r_{\rm surface} = %.2f$' %r, fontsize=20)




def plotMagField(datDir=datDir,planet="earth",r=1,levels=30,proj='moll',cmap='RdBu_r'):

    p2D, th2D, Br, dum1,dum2 = getBr(datDir=datDir,planet=planet,r=r)
    plt.figure(figsize=(12,6.75))
    plotB(p2D,th2D,Br,planet=planet,levels=levels,proj=proj,cmap=cmap,r=r)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import scipy.special as sp
from copy import deepcopy
from importlib_resources import files

def gen_idx(lmax):

    '''
    Generate index to convert from (l,m) to [idx]
    '''

    idx = np.zeros([lmax+1,lmax+1])
    count = 0

    for l in range(lmax+1):
        for m in range(l+1):

            idx[l,m] = count
            count += 1

    return np.int32(idx)

def get_data(datDir,planet="earth"):

    datfile = datDir.joinpath(planet+'.dat')

    if planet == "earth":
        datfile = datDir.joinpath('IGRF13.dat')
        dat = np.loadtxt(datfile,usecols=[1,2,-2])
        gh  = np.genfromtxt(datfile,usecols=[0],dtype='str')

        mask = gh == 'g'
        gDat = dat[mask,:]
        g   = gDat[:,-1]
        gl  = gDat[:,0]

        mask = gh == 'h'
        hDat = dat[mask,:]
        h   = hDat[:,-1]
        hm  = hDat[:,1]
        hIdx = np.where(hm == 1.)[0]
        h   = np.insert(h,hIdx,0.)

        lmax = np.int32(gl.max())

    elif planet in ["mercury","saturn"]:
        dat = np.loadtxt(datfile,usecols=[3])
        g   = dat.flatten()
        lmax = len(g)
        h = np.zeros_like(g)

    elif planet == "jupiter":
        dat = np.loadtxt(datfile)
        l_dat = dat[:,-2]
        ghlm = dat[:,1]

        lmax = np.int32(l_dat.max())

        g = []
        h = []

        ########################
        # Separate glm and hlm
        ########################

        for i in range(1,lmax+1):
            mask = l_dat == i
            n = len(l_dat[mask])
            half = int(n/2)

            g.append(ghlm[mask][:half+1])
            h.append(np.concatenate([[0.],ghlm[mask][half+1:]]))

        g  = np.concatenate(g)
        h  = np.concatenate(h)

    elif planet in ['uranus','neptune','ganymede']:
        dat = np.loadtxt(datfile,usecols=[1,2,3])
        gh  = np.genfromtxt(datfile,usecols=[0],dtype='str')

        mask = gh == 'g'
        gDat = dat[mask,:]
        g   = gDat[:,-1]
        gl  = gDat[:,0]

        mask = gh == 'h'
        hDat = dat[mask,:]
        h   = hDat[:,-1]

        gl = np.int32(gl)

        lmax = np.int32(gl.max())

    # Insert (0,0) -> 0 for less confusion

    g = np.insert(g,0,0.)
    h = np.insert(h,0,0.)

    if planet in ['mercury','saturn']:
        idx = np.arange(lmax+1)
    else:
        idx = gen_idx(lmax)

    return g,h,lmax,idx

def get_grid(nphi=256,ntheta=128):

    phi    = np.linspace(0.,2*np.pi,nphi)
    x,w    = sp.roots_legendre(ntheta)
    theta  = np.sort(np.arccos(x))

    p2D  = np.zeros([nphi, ntheta])
    th2D = np.zeros([nphi, ntheta])

    for i in range(nphi):
        p2D[i,:] = phi[i]

    for j in range(ntheta):
        th2D[:,j] = theta[j]

    return p2D, th2D

def gen_arr(lmax, l1,m1,mode='g'):

    '''
    Generate Gauss coefficient array for testing
    purposes
    '''

    idx = np.zeros([lmax+1,lmax+1])
    lArr = []
    mArr = []

    count = 0

    glm = []
    hlm = []

    for l in range(lmax+1):
        for m in range(l+1):

            if l in l1 and m in m1:
                if mode == 'g' or mode == 'gh':
                    glm.append(1.)
                else:
                    glm.append(0.)
                if mode == 'h' or mode == 'gh':
                    hlm.append(1.)
                else:
                    hlm.append(0.)
            else:
                glm.append(0.)
                hlm.append(0.)

            idx[l,m] = count
            lArr.append(l)
            mArr.append(m)

            count += 1

    glm  = np.array(glm)
    hlm  = np.array(hlm)
    lArr = np.array(lArr)
    mArr = np.array(mArr)
    idx =  np.int32(idx)

    return glm, hlm, lArr, mArr, idx

def getB(lmax,glm,hlm,idx,r,p2D,th2D,planet="earth"):

    Br = np.zeros_like(p2D)

    for l in range(1,lmax+1):
        for m in range(l+1):
            ylm = sp.sph_harm(m, l, p2D, th2D)

            # Include Condon-Shortley Phase for Earth but not other planets
            # Scipy sph_harm has the phase included by default

            if planet in ["earth"]:
                fac_m = 1.
            else:
                fac_m = (-1)**m

            if m != 0:
                fac_m *= np.sqrt(2)

            fac = fac_m * (l+1) * r**(-l-2) * np.sqrt((4.*np.pi)/(2*l+1))

            G = glm[idx[l,m]] * np.real(ylm)
            H = hlm[idx[l,m]] * np.imag(ylm)

            Br +=   np.real(fac * (G + H))

    return Br

def getBm0(lmax,g,r,p2D,th2D):

    Br = np.zeros_like(p2D)

    for l in range(1,lmax+1):
        ylm = sp.sph_harm(0, l, p2D, th2D)
        fac = (l + 1) * r**(-l-2) * np.sqrt((4.*np.pi)/(2*l+1))

        Br += fac * g[l] * np.real(ylm)

    return Br

def get_spec(glm,hlm,idx,lmax,planet='earth',r=1):
    E = np.zeros(lmax+1)

    if planet in ['mercury','saturn']:
        for l in range(1,lmax+1):
            E[l] = (l+1) * r**(-2*l-4) *(np.abs(glm[l])**2 + np.abs(hlm[l])**2)
        emag_10 = E[1]
    else:
        for l in range(1,lmax+1):
            for m in range(l+1):
                E[l] += (l+1) * r**(-2*l-4) *(np.abs(glm[idx[l,m]])**2 + np.abs(hlm[idx[l,m]])**2)

        emag_10 = 2 * r**(-2*l-4)* np.abs(glm[idx[1,0]])**2
    return E, emag_10

def filt_Gauss(glm,hlm,lmax,idx,larr=None,marr=None,lCutMin=0,lCutMax=None,mmin=0,mmax=None):

    glm_filt = deepcopy(glm)
    hlm_filt = deepcopy(hlm)

    if lCutMax is None:
        lCutMax = lmax
    if mmax is None:
        mmax = lmax

    if larr is not None:
        if max(larr) > lmax:
            print("Error! Values in filter array must be <= lmax=%d" %lmax)
        else:
            for ell in range(lmax+1):
                if ell not in larr:
                    glm_filt[idx[ell,:]] = 0.
                    hlm_filt[idx[ell,:]] = 0.
    else:
        if lCutMax > lmax or lCutMin > lmax:
            print("Error! lCutMin/lCutMax must be <= lmax = %d" %lmax)
        else:
            for ell in range(lCutMin):
                    glm_filt[idx[ell,:]] = 0.
                    hlm_filt[idx[ell,:]] = 0.
            for ell in range(lCutMax,lmax+1):
                    glm_filt[idx[ell,:]] = 0.
                    hlm_filt[idx[ell,:]] = 0.

    if marr is not None:
        if max(marr) > lmax:
            print("Error! Values in filter array must be <= lmax=%d" %lmax)
        else:
            for m in range(lmax+1):
                if m not in marr:
                    glm_filt[idx[:,m]] = 0.
                    hlm_filt[idx[:,m]] = 0.
    else:
        if mmin > lmax or mmax > lmax:
            print("Error! mmin/mmax must be <= lmax = %d" %lmax)
        else:
            for m in range(mmin):
                    glm_filt[idx[:,m]] = 0.
                    hlm_filt[idx[:,m]] = 0.
            for m in range(mmax+1,lmax+1):
                    glm_filt[idx[:,m]] = 0.
                    hlm_filt[idx[:,m]] = 0.

    return glm_filt,hlm_filt

def filt_Gaussm0(glm,hlm,lmax,larr=None,lCutMin=0,lCutMax=None):

    glm_filt = deepcopy(glm)
    hlm_filt = deepcopy(hlm)

    if lCutMax is None:
        lCutMax = lmax

    if larr is not None:
        if max(larr) > lmax:
            print("Error! Values in filter array must be <= lmax=%d" %lmax)
        else:
            for ell in range(lmax+1):
                if ell not in larr:
                    glm_filt[ell] = 0.
                    hlm_filt[ell] = 0.
    else:
        if lCutMax > lmax or lCutMin > lmax:
            print("Error! lCutMin/lCutMax must be <= lmax = %d" %lmax)
        else:
            for ell in range(lCutMin):
                    glm_filt[ell] = 0.
                    hlm_filt[ell] = 0.
            for ell in range(lCutMax,lmax+1):
                    glm_filt[ell] = 0.
                    hlm_filt[ell] = 0.

    return glm_filt,hlm_filt


def sphInt(f,g,phi,th2D,theta):

    thetaInt = np.trapz(f * g * np.sin(th2D),theta,axis=1)
    phiInt = np.trapz(thetaInt,phi)

    return phiInt


def getGauss(lmax,Br,r,phi,theta,th2D,p2D):

    '''
    Get Gauss coefficients from a surface field
    '''
    glm = []
    hlm = []

    comp = complex(0,1)

    for l in range(0,lmax+1):
        for m in range(0,l+1):

            ylm = (-1)**m * sp.sph_harm(m, l, p2D, th2D)

            ylm_conj = np.conjugate(ylm)

            fac = r**(l+2)/(l+1)

            if m==0:
                fac *= 0.5

            I1 = sphInt(Br,ylm,phi,th2D,theta)
            I2 = sphInt(Br,ylm_conj,phi,th2D,theta)

            g = fac * (I2 + I1)
            h = comp * fac * (I2 - I1)

            glm.append(g)
            hlm.append(h)

    glm = np.array(glm)
    hlm = np.array(hlm)

    return glm, hlm
